#include <ppsi/ppsi.h>

/* ext-whiterabbit must offer its own hooks */

int wrTmoIdx=0; /* TimeOut Index */

static int wr_init(struct pp_instance *ppi, void *buf, int len)
{
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);

	if ( wrTmoIdx==0)
		wrTmoIdx=pp_timeout_get_timer(ppi,"WR_EXT_0",TO_RAND_NONE, TMO_CF_INSTANCE_DEPENDENT);

	wr_reset_process(ppi,WR_ROLE_NONE);

#ifdef CONFIG_ABSCAL
        /* absolute calibration only exists in arch-wrpc, so far */
        extern int ptp_mode;
        if (ptp_mode == 4 /* WRC_MODE_ABSCAL */)
                ppi->next_state = WRS_WR_LINK_ON;
#endif

	return 0;
}

/* open hook called only for each WR pp_instances */
static int wr_open(struct pp_instance *ppi, struct pp_runtime_opts *rt_opts)
{
	pp_diag(NULL, ext, 2, "hook: %s\n", __func__);

	if ( is_slaveOnly(DSDEF(ppi)) ||
			( is_externalPortConfigurationEnabled(DSDEF(ppi)) &&
					ppi->externalPortConfigurationPortDS.desiredState==PPS_SLAVE)
	   ){
		WR_DSPOR(ppi)->wrConfig = WR_S_ONLY;
	} else {
		WR_DSPOR(ppi)->wrConfig = ( ppi->portDS->masterOnly ||
				( is_externalPortConfigurationEnabled(DSDEF(ppi)) &&
						ppi->externalPortConfigurationPortDS.desiredState==PPS_MASTER)) ?
								WR_M_ONLY :
								WR_M_AND_S;
	}
	return 0;
}

static int wr_handle_resp(struct pp_instance *ppi)
{
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);

	/* This correction_field we received is already part of t4 */
	if ( ppi->ext_enabled ) {
		wr_servo_got_resp(ppi);
	}
	else {
		pp_servo_got_resp(ppi,OPTS(ppi)->ptpFallbackPpsGen);
	}
	return 0;
}


static int  wr_sync_followup(struct pp_instance *ppi) {

	if ( ppi->ext_enabled ) {
		wr_servo_got_sync(ppi);
	}
	else {
		pp_servo_got_sync(ppi,OPTS(ppi)->ptpFallbackPpsGen);
	}
	return 1; /* the caller returns too */
}

static int wr_handle_sync(struct pp_instance *ppi)
{
	/* This handle is called in case of one step clock */
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);
	return wr_sync_followup(ppi);
}

static int wr_handle_followup(struct pp_instance *ppi)
{
	/* This handle is called in case of two step clock */
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);
	return wr_sync_followup(ppi);
}

static int wr_handle_presp(struct pp_instance *ppi)
{
	/* FIXME: verify that last-received cField is already accounted for */
	if ( ppi->ext_enabled )
		wr_servo_got_presp(ppi);
	else
		pp_servo_got_presp(ppi);
	return 0;
}

/* Used only in MASTER state to re-enable the extension if needed */
static int wr_handle_signaling(struct pp_instance *ppi, void *buf, int len) {
	if ( !ppi->ext_enabled && ppi->state==PPS_MASTER ) {
		/* Check if WR_PRESENT message is received */
		MsgSignaling wrsig_msg;
		Enumeration16 msgId;

		msg_unpack_wrsig(ppi, buf, &wrsig_msg, &msgId);

		if (msgId==SLAVE_PRESENT) {
			/* Re-enable the extension */
			lstate_enable_extension(ppi);
		}
	}
	return 0;
}

static int wr_pack_announce(struct pp_instance *ppi)
{
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);
	/* Even though the extension is disable we continue to send WR TLV */
	if ( WR_DSPOR(ppi)->wrConfig != WR_S_ONLY) {
		msg_pack_announce_wr_tlv(ppi);
		return WR_ANNOUNCE_LENGTH;
	}
	return PP_ANNOUNCE_LENGTH;
}

static void wr_unpack_announce(struct pp_instance *ppi,void *buf, MsgAnnounce *ann)
{
	int msg_len = ntohs(*(UInteger16 *) (buf + 2));

	pp_diag(NULL, ext, 2, "hook: %s\n", __func__);
	if (msg_len >= WR_ANNOUNCE_LENGTH) {
		UInteger16 wr_flags;
		Boolean parentIsWRnode;
		struct wr_dsport *wrp = WR_DSPOR(ppi);
		MsgHeader *hdr = &ppi->received_ptp_header;
		struct PortIdentity *pid = &hdr->sourcePortIdentity;

		msg_unpack_announce_wr_tlv(buf, ann, &wr_flags);
		parentIsWRnode=(wr_flags & WR_NODE_MODE)!=NON_WR;

		// Check if we need to re-enable the extension
		if ( !ppi->ext_enabled ) {
			// The extension must be re-enabled only if :
			// - The parent is a WR node
			// - ptp state=(slave|uncalibrated|listening)
			// - Same parent port ID but with a not continuous sequence ID (With a margin of 1)
			// - The port identity is different
			if ( parentIsWRnode &&
					(ppi->state==PPS_SLAVE ||
					ppi->state==PPS_UNCALIBRATED ||
					ppi->state==PPS_LISTENING))  {

				Boolean samePid=!bmc_pidcmp(pid, &wrp->parentAnnPortIdentity);
				if ( !samePid  ||
						(samePid &&
								(hdr->sequenceId!=wrp->parentAnnSequenceId+1 &&
								hdr->sequenceId!=wrp->parentAnnSequenceId+2)
								)) {
					lstate_enable_extension(ppi);
				}
			} else {
				parentIsWRnode=FALSE;
			}
		}

		memcpy(&wrp->parentAnnPortIdentity,pid,sizeof(struct PortIdentity));
		wrp->parentAnnSequenceId=hdr->sequenceId;

		/* Update the WR parent state */
		if ( !parentIsWRnode  )
			/* Not a WR node */
			wr_flags=0; /* Forget all bits. They are not relevant */
		wrp->parentIsWRnode = parentIsWRnode;
		wrp->parentWrModeOn = (wr_flags & WR_IS_WR_MODE) != 0;
		wrp->parentCalibrated =(wr_flags & WR_IS_CALIBRATED) != 0;
		wrp->parentWrConfig = wr_flags & WR_NODE_MODE;
	}
}

static void wr_state_change(struct pp_instance *ppi)
{
	struct wr_dsport *wrp = WR_DSPOR(ppi);
	
	pp_diag(ppi, ext, 2, "hook: %s\n", __func__);
	if ( ppi->ext_enabled ) {
		if ( wrp->wrModeOn &&
				((ppi->state == PPS_SLAVE) ||
						(ppi->state == PPS_MASTER))) {
			/* if we are leaving the MASTER or SLAVE state */
			wr_reset_process(ppi,WR_ROLE_NONE);
			lstate_set_link_pdetection(ppi);
			if (ppi->state == PPS_SLAVE)
				//* We are leaving SLAVE state
				WRH_OPER()->locking_reset(ppi);
		} else {
			switch (ppi->next_state) {
			case PPS_MASTER : /* Enter in MASTER_STATE */
				wrp->next_state=WRS_IDLE;
				wrp->wrMode=WR_MASTER;
				break;
			case PPS_UNCALIBRATED : /* Enter in UNCALIBRATED state */
				wrp->next_state=WRS_PRESENT;
				wrp->wrMode=WR_SLAVE;
				lstate_set_link_pdetection(ppi);
				break;
			case PPS_SLAVE : /* Enter in SLAVE state */
				if ( wrp->wrModeOn &&  wrp->parentWrModeOn ) {
					lstate_set_link_established(ppi);
				}
				break;
			case PPS_LISTENING : /* Enter in LISTENING state */
				wr_reset_process(ppi,WR_ROLE_NONE);
				break;
			}
		}

		if ( ppi->state==PPS_SLAVE && ppi->next_state!=PPS_UNCALIBRATED ) {
			/* Leave SLAVE state : We must stop the PPS generation */
			WRH_OPER()->enable_timing_output(GLBS(ppi),0);
		}
	} else
		wr_reset_process(ppi,WR_ROLE_NONE);

}

int wr_ready_for_slave(struct pp_instance *ppi) {
	if ( ppi->ext_enabled ) {
		struct wr_dsport *wrp = WR_DSPOR(ppi);
		return wrp->wrModeOn && wrp->parentWrModeOn && wrp->state == WRS_IDLE;
	} else
		return 1; /* Allow to go to slave state as we will not do a WR handshake */
}


static int wr_require_precise_timestamp(struct pp_instance *ppi) {
	return ppi->ext_enabled  && WR_DSPOR(ppi)->wrModeOn;
}

static int wr_get_tmo_lstate_detection(struct pp_instance *ppi) {
	/* The WR protocol detection comes :
	 * - for a SLAVE: from the announce message (WR TLV)
	 * - for a MASTER: from the reception of the WR_PRESENT message
	 * To cover this 2 cases we will use 2 times the announce receipt time-out
	 */
	return is_externalPortConfigurationEnabled(DSDEF(ppi)) ?
			20000 : /* 20s: externalPortConfiguration enable means no ANN_RECEIPT timeout */
			pp_timeout_get(ppi,PP_TO_ANN_RECEIPT)<<3;
}

static TimeInterval wr_get_latency (struct pp_instance *ppi) {
	return 0;
}

/* WR extension is not compliant with the standard concerning the contents of the correction fields */
static int wr_is_correction_field_compliant (struct pp_instance *ppi) {
	return 0;
}
struct pp_ext_hooks wr_ext_hooks = {
	.init = wr_init,
	.open = wr_open,
	.handle_resp = wr_handle_resp,
	.handle_sync = wr_handle_sync,
	.handle_signaling=wr_handle_signaling,
	.handle_followup = wr_handle_followup,
	.ready_for_slave = wr_ready_for_slave,
	.run_ext_state_machine = wr_run_state_machine,
#if CONFIG_HAS_P2P
	.handle_presp = wr_handle_presp,
#endif
	.pack_announce = wr_pack_announce,
	.unpack_announce = wr_unpack_announce,
	.state_change = wr_state_change,
	.servo_reset= wr_servo_reset,
	.require_precise_timestamp=wr_require_precise_timestamp,
	.get_tmo_lstate_detection=wr_get_tmo_lstate_detection,
	.get_ingress_latency=wr_get_latency,
	.get_egress_latency=wr_get_latency,
	.is_correction_field_compliant=wr_is_correction_field_compliant,
};
